"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.GlobalDeclarationInstantiation = GlobalDeclarationInstantiation;

exports.default = function (ast, strictCode, env, realm) {
  strictCode = (0, _strict2.default)(ast);

  GlobalDeclarationInstantiation(realm, ast, env, strictCode);

  var context = realm.getRunningContext();
  var val = void 0;

  var _iteratorNormalCompletion11 = true;
  var _didIteratorError11 = false;
  var _iteratorError11 = undefined;

  try {
    for (var _iterator11 = ast.body[Symbol.iterator](), _step11; !(_iteratorNormalCompletion11 = (_step11 = _iterator11.next()).done); _iteratorNormalCompletion11 = true) {
      var node = _step11.value;

      if (node.type !== "FunctionDeclaration") {
        var potentialVal = env.evaluateCompletion(node, strictCode);
        if (potentialVal instanceof _completions.Completion) {
          if (!realm.useAbstractInterpretation) throw potentialVal;
          // todo: emit code to throw exeptions at runtime
          throw new _errors.FatalError();
        }
        if (context.savedCompletion !== undefined) {
          // todo: emit conditional code to throw exeptions at runtime
          throw new _errors.FatalError();
        }
        if (!(potentialVal instanceof _index.EmptyValue)) val = potentialVal;
      }
    }
  } catch (err) {
    _didIteratorError11 = true;
    _iteratorError11 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion11 && _iterator11.return) {
        _iterator11.return();
      }
    } finally {
      if (_didIteratorError11) {
        throw _iteratorError11;
      }
    }
  }

  var directives = ast.directives;
  if (!val && directives && directives.length) {
    var directive = directives[directives.length - 1];
    val = env.evaluate(directive, strictCode);
  }

  (0, _invariant2.default)(val === undefined || val instanceof _index.Value);
  return val || realm.intrinsics.empty;
};

var _completions = require("../completions.js");

var _errors = require("../errors.js");

var _index = require("../values/index.js");

var _environment = require("../environment.js");

var _function = require("../methods/function.js");

var _index2 = require("../methods/index.js");

var _strict = require("../utils/strict.js");

var _strict2 = _interopRequireDefault(_strict);

var _invariant = require("../invariant.js");

var _invariant2 = _interopRequireDefault(_invariant);

var _traverseFast = require("../utils/traverse-fast.js");

var _traverseFast2 = _interopRequireDefault(_traverseFast);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// ECMA262 15.1.11
function GlobalDeclarationInstantiation(realm, ast, env, strictCode) {
  realm.getRunningContext().isStrict = realm.isStrict = strictCode;

  // 1. Let envRec be env's EnvironmentRecord.
  var envRec = env.environmentRecord;

  // 2. Assert: envRec is a global Environment Record.
  (0, _invariant2.default)(envRec instanceof _environment.GlobalEnvironmentRecord, "expected global environment record");

  // 3. Let lexNames be the LexicallyDeclaredNames of script.
  var lexNames = [];

  // 4. Let varNames be the VarDeclaredNames of script.
  var varNames = [];

  (0, _traverseFast2.default)(ast, function (node) {
    if (node.type === "VariableDeclaration") {
      if (node.kind === "var") {
        varNames = varNames.concat((0, _index2.BoundNames)(realm, node));
      } else {
        lexNames = lexNames.concat((0, _index2.BoundNames)(realm, node));
      }
    } else if (node.type === "FunctionExpression" || node.type === "FunctionDeclaration") {
      return true;
    }
    return false;
  });

  // 5. For each name in lexNames, do
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = lexNames[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var name = _step.value;

      // a. If envRec.HasVarDeclaration(name) is true, throw a SyntaxError exception.
      if (envRec.HasVarDeclaration(name)) {
        throw realm.createErrorThrowCompletion(realm.intrinsics.SyntaxError, name + " already declared with var");
      }

      // b. If envRec.HasLexicalDeclaration(name) is true, throw a SyntaxError exception.
      if (envRec.HasLexicalDeclaration(name)) {
        throw realm.createErrorThrowCompletion(realm.intrinsics.SyntaxError, name + " already declared with let or const");
      }

      // c. Let hasRestrictedGlobal be ? envRec.HasRestrictedGlobalProperty(name).
      var hasRestrictedGlobal = envRec.HasRestrictedGlobalProperty(name);

      // d. If hasRestrictedGlobal is true, throw a SyntaxError exception.
      if (hasRestrictedGlobal) {
        throw realm.createErrorThrowCompletion(realm.intrinsics.SyntaxError, name + " global object is restricted");
      }
    }

    // 6. For each name in varNames, do
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  var _iteratorNormalCompletion2 = true;
  var _didIteratorError2 = false;
  var _iteratorError2 = undefined;

  try {
    for (var _iterator2 = varNames[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
      var _name = _step2.value;

      // a. If envRec.HasLexicalDeclaration(name) is true, throw a SyntaxError exception.
      if (envRec.HasLexicalDeclaration(_name)) {
        throw realm.createErrorThrowCompletion(realm.intrinsics.SyntaxError, _name + " already declared with let or const");
      }
    }

    // 7. Let varDeclarations be the VarScopedDeclarations of script.
  } catch (err) {
    _didIteratorError2 = true;
    _iteratorError2 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion2 && _iterator2.return) {
        _iterator2.return();
      }
    } finally {
      if (_didIteratorError2) {
        throw _iteratorError2;
      }
    }
  }

  var varDeclarations = (0, _function.FindVarScopedDeclarations)(ast);

  // 8. Let functionsToInitialize be a new empty List.
  var functionsToInitialize = [];

  // 9. Let declaredFunctionNames be a new empty List.
  var declaredFunctionNames = [];

  // 10. For each d in varDeclarations, in reverse list order do
  var _iteratorNormalCompletion3 = true;
  var _didIteratorError3 = false;
  var _iteratorError3 = undefined;

  try {
    for (var _iterator3 = varDeclarations.reverse()[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
      var d = _step3.value;

      // a. If d is neither a VariableDeclaration or a ForBinding, then
      if (d.type !== "VariableDeclaration") {
        // i. Assert: d is either a FunctionDeclaration or a GeneratorDeclaration.
        (0, _invariant2.default)(d.type === "FunctionDeclaration", "expected function");

        // ii. NOTE If there are multiple FunctionDeclarations for the same name, the last declaration is used.

        // iii. Let fn be the sole element of the BoundNames of d.
        var fn = (0, _index2.BoundNames)(realm, d)[0];

        // iv. If fn is not an element of declaredFunctionNames, then
        if (declaredFunctionNames.indexOf(fn) < 0) {
          // 1. Let fnDefinable be ? envRec.CanDeclareGlobalFunction(fn).
          var fnDefinable = envRec.CanDeclareGlobalFunction(fn);

          // 2. If fnDefinable is false, throw a TypeError exception.
          if (!fnDefinable) {
            throw realm.createErrorThrowCompletion(realm.intrinsics.TypeError, fn + ": global function declarations are not allowed");
          }

          // 3. Append fn to declaredFunctionNames.
          declaredFunctionNames.push(fn);

          // 4. Insert d as the first element of functionsToInitialize.
          functionsToInitialize.unshift(d);
        }
      }
    }

    // 11. Let declaredVarNames be a new empty List.
  } catch (err) {
    _didIteratorError3 = true;
    _iteratorError3 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion3 && _iterator3.return) {
        _iterator3.return();
      }
    } finally {
      if (_didIteratorError3) {
        throw _iteratorError3;
      }
    }
  }

  var declaredVarNames = [];

  // 12. For each d in varDeclarations, do
  var _iteratorNormalCompletion4 = true;
  var _didIteratorError4 = false;
  var _iteratorError4 = undefined;

  try {
    for (var _iterator4 = varDeclarations[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
      var _d = _step4.value;

      // a. If d is a VariableDeclaration or a ForBinding, then
      if (_d.type === "VariableDeclaration") {
        // i. For each String vn in the BoundNames of d, do
        var _iteratorNormalCompletion9 = true;
        var _didIteratorError9 = false;
        var _iteratorError9 = undefined;

        try {
          for (var _iterator9 = (0, _index2.BoundNames)(realm, _d)[Symbol.iterator](), _step9; !(_iteratorNormalCompletion9 = (_step9 = _iterator9.next()).done); _iteratorNormalCompletion9 = true) {
            var _vn = _step9.value;

            // ii. If vn is not an element of declaredFunctionNames, then
            if (declaredFunctionNames.indexOf(_vn) < 0) {
              // 1. Let vnDefinable be ? envRec.CanDeclareGlobalVar(vn).
              var vnDefinable = envRec.CanDeclareGlobalVar(_vn);

              // 2. If vnDefinable is false, throw a TypeError exception.
              if (!vnDefinable) {
                throw realm.createErrorThrowCompletion(realm.intrinsics.TypeError, _vn + ": global variable declarations are not allowed");
              }

              // 3. If vn is not an element of declaredVarNames, then
              if (declaredVarNames.indexOf(_vn) < 0) {
                // a. Append vn to declaredVarNames.
                declaredVarNames.push(_vn);
              }
            }
          }
        } catch (err) {
          _didIteratorError9 = true;
          _iteratorError9 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion9 && _iterator9.return) {
              _iterator9.return();
            }
          } finally {
            if (_didIteratorError9) {
              throw _iteratorError9;
            }
          }
        }
      }
    }

    // 13. NOTE: No abnormal terminations occur after this algorithm step if the global object is an ordinary object. However, if the global object is a Proxy exotic object it may exhibit behaviours that cause abnormal terminations in some of the following steps.

    // 14. NOTE: Annex B.3.3.2 adds additional steps at this point.

    // 15. Let lexDeclarations be the LexicallyScopedDeclarations of script.
  } catch (err) {
    _didIteratorError4 = true;
    _iteratorError4 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion4 && _iterator4.return) {
        _iterator4.return();
      }
    } finally {
      if (_didIteratorError4) {
        throw _iteratorError4;
      }
    }
  }

  var lexDeclarations = [];
  var _iteratorNormalCompletion5 = true;
  var _didIteratorError5 = false;
  var _iteratorError5 = undefined;

  try {
    for (var _iterator5 = ast.body[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
      var s = _step5.value;

      if (s.type === "VariableDeclaration" && s.kind !== "var") {
        lexDeclarations.push(s);
      }
    }

    // 16. For each element d in lexDeclarations do
  } catch (err) {
    _didIteratorError5 = true;
    _iteratorError5 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion5 && _iterator5.return) {
        _iterator5.return();
      }
    } finally {
      if (_didIteratorError5) {
        throw _iteratorError5;
      }
    }
  }

  var _iteratorNormalCompletion6 = true;
  var _didIteratorError6 = false;
  var _iteratorError6 = undefined;

  try {
    for (var _iterator6 = lexDeclarations[Symbol.iterator](), _step6; !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
      var _d2 = _step6.value;

      // a. NOTE Lexically declared names are only instantiated here but not initialized.

      // b. For each element dn of the BoundNames of d do
      var _iteratorNormalCompletion10 = true;
      var _didIteratorError10 = false;
      var _iteratorError10 = undefined;

      try {
        for (var _iterator10 = (0, _index2.BoundNames)(realm, _d2)[Symbol.iterator](), _step10; !(_iteratorNormalCompletion10 = (_step10 = _iterator10.next()).done); _iteratorNormalCompletion10 = true) {
          var dn = _step10.value;

          // i. If IsConstantDeclaration of d is true, then
          if (_d2.kind === "const") {
            // 1. Perform ? envRec.CreateImmutableBinding(dn, true).
            envRec.CreateImmutableBinding(dn, true);
          } else {
            // ii. Else,
            // 1. Perform ? envRec.CreateMutableBinding(dn, false).
            envRec.CreateMutableBinding(dn, false);
          }
        }
      } catch (err) {
        _didIteratorError10 = true;
        _iteratorError10 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion10 && _iterator10.return) {
            _iterator10.return();
          }
        } finally {
          if (_didIteratorError10) {
            throw _iteratorError10;
          }
        }
      }
    }

    // 17. For each production f in functionsToInitialize, do
  } catch (err) {
    _didIteratorError6 = true;
    _iteratorError6 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion6 && _iterator6.return) {
        _iterator6.return();
      }
    } finally {
      if (_didIteratorError6) {
        throw _iteratorError6;
      }
    }
  }

  var _iteratorNormalCompletion7 = true;
  var _didIteratorError7 = false;
  var _iteratorError7 = undefined;

  try {
    for (var _iterator7 = functionsToInitialize[Symbol.iterator](), _step7; !(_iteratorNormalCompletion7 = (_step7 = _iterator7.next()).done); _iteratorNormalCompletion7 = true) {
      var f = _step7.value;

      // a. Let fn be the sole element of the BoundNames of f.
      var fn = (0, _index2.BoundNames)(realm, f)[0];

      // b. Let fo be the result of performing InstantiateFunctionObject for f with argument env.
      var fo = env.evaluate(f, strictCode);
      (0, _invariant2.default)(fo instanceof _index.Value);

      // c. Perform ? envRec.CreateGlobalFunctionBinding(fn, fo, false).
      envRec.CreateGlobalFunctionBinding(fn, fo, false);
    }

    // 18. For each String vn in declaredVarNames, in list order do
  } catch (err) {
    _didIteratorError7 = true;
    _iteratorError7 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion7 && _iterator7.return) {
        _iterator7.return();
      }
    } finally {
      if (_didIteratorError7) {
        throw _iteratorError7;
      }
    }
  }

  var _iteratorNormalCompletion8 = true;
  var _didIteratorError8 = false;
  var _iteratorError8 = undefined;

  try {
    for (var _iterator8 = declaredVarNames[Symbol.iterator](), _step8; !(_iteratorNormalCompletion8 = (_step8 = _iterator8.next()).done); _iteratorNormalCompletion8 = true) {
      var vn = _step8.value;

      // a. Perform ? envRec.CreateGlobalVarBinding(vn, false).
      envRec.CreateGlobalVarBinding(vn, false);
    }

    // 19. Return NormalCompletion(empty).
  } catch (err) {
    _didIteratorError8 = true;
    _iteratorError8 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion8 && _iterator8.return) {
        _iterator8.return();
      }
    } finally {
      if (_didIteratorError8) {
        throw _iteratorError8;
      }
    }
  }

  return realm.intrinsics.empty;
} /**
   * Copyright (c) 2017-present, Facebook, Inc.
   * All rights reserved.
   *
   * This source code is licensed under the BSD-style license found in the
   * LICENSE file in the root directory of this source tree. An additional grant
   * of patent rights can be found in the PATENTS file in the same directory.
   */
//# sourceMappingURL=Program.js.map