"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }(); /**
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          * Copyright (c) 2017-present, Facebook, Inc.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          * All rights reserved.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          * This source code is licensed under the BSD-style license found in the
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          * LICENSE file in the root directory of this source tree. An additional grant
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          * of patent rights can be found in the PATENTS file in the same directory.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          */

exports.default = simplifyAbstractValue;

var _errors = require("../errors.js");

var _index = require("../domains/index.js");

var _invariant = require("../invariant.js");

var _invariant2 = _interopRequireDefault(_invariant);

var _index2 = require("../methods/index.js");

var _realm = require("../realm.js");

var _index3 = require("../values/index.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function simplifyAbstractValue(realm, value) {
  var savedHandler = realm.errorHandler;
  var savedIsReadOnly = realm.isReadOnly;
  realm.isReadOnly = true;
  try {
    realm.errorHandler = function () {
      throw new _errors.FatalError();
    };
    return simplify(realm, value);
  } catch (e) {
    return value;
  } finally {
    realm.errorHandler = savedHandler;
    realm.isReadOnly = savedIsReadOnly;
  }
}

function simplify(realm, value) {
  if (value instanceof _index3.ConcreteValue) return value;
  (0, _invariant2.default)(value instanceof _index3.AbstractValue);
  switch (value.kind) {
    case "!":
      return negate(realm, value.args[0]);
    case "||":
    case "&&":
      var x = simplify(realm, value.args[0]);
      var y = simplify(realm, value.args[1]);
      if (x instanceof _index3.AbstractValue && x.equals(y)) return x;
      return _index3.AbstractValue.createFromLogicalOp(realm, value.kind, x, y, value.expressionLocation);
    case "==":
    case "!=":
    case "===":
    case "!==":
      return simplifyEquality(realm, value);
    default:
      return value;
  }
}

function simplifyEquality(realm, equality) {
  var op = equality.kind;

  var _equality$args = _slicedToArray(equality.args, 2),
      x = _equality$args[0],
      y = _equality$args[1];

  if (x instanceof _index3.ConcreteValue) {
    ;
    var _ref = [y, x];
    x = _ref[0];
    y = _ref[1];
  }if (x instanceof _index3.AbstractValue && x.kind === "conditional" && (!y.mightNotBeUndefined() || !y.mightNotBeNull())) {
    // try to simplify "(cond ? xx : xy) op undefined/null" to just "cond" or "!cond"
    var _x$args = _slicedToArray(x.args, 3),
        cond = _x$args[0],
        xx = _x$args[1],
        xy = _x$args[2];

    (0, _invariant2.default)(cond instanceof _index3.AbstractValue); // otherwise the the conditional should not have been created
    if (op === "===" || op === "!==") {
      // if xx === undefined && xy !== undefined then cond <=> x === undefined
      if (!y.mightNotBeUndefined() && !xx.mightNotBeUndefined() && !xy.mightBeUndefined()) return op === "===" ? makeBoolean(realm, cond) : negate(realm, cond);
      // if xx !== undefined && xy === undefined then !cond <=> x === undefined
      if (!y.mightNotBeUndefined() && !xx.mightBeUndefined() && !xy.mightNotBeUndefined()) return op === "===" ? negate(realm, cond) : makeBoolean(realm, cond);
      // if xx === null && xy !== null then cond <=> x === null
      if (!y.mightNotBeNull() && !xx.mightNotBeNull() && !xy.mightBeNull()) return op === "===" ? makeBoolean(realm, cond) : negate(realm, cond);
      // if xx !== null && xy === null then !cond <=> x === null
      if (!y.mightNotBeNull() && !xx.mightBeNull() && !xy.mightNotBeNull()) return op === "===" ? negate(realm, cond) : makeBoolean(realm, cond);
    } else {
      (0, _invariant2.default)(op === "==" || op === "!=");
      // if xx cannot be undefined/null and xy is undefined/null then !cond <=> x == undefined/null
      if (!xx.mightBeUndefined() && !xx.mightBeNull() && (!xy.mightNotBeUndefined() || !xy.mightNotBeNull())) return op === "==" ? negate(realm, cond) : makeBoolean(realm, cond);
      // if xx is undefined/null and xy cannot be undefined/null then cond <=> x == undefined/null
      if ((!xx.mightNotBeUndefined() || !xx.mightNotBeNull()) && !xy.mightBeUndefined() && !xy.mightBeNull()) return op === "==" ? makeBoolean(realm, cond) : negate(realm, cond);
    }
  }
  return equality;
}

function makeBoolean(realm, value) {
  if (value.getType() === _index3.BooleanValue) return value;
  if (value instanceof _index3.ConcreteValue) return new _index3.BooleanValue(realm, (0, _index2.ToBoolean)(realm, value));
  (0, _invariant2.default)(value instanceof _index3.AbstractValue);
  var v = _index3.AbstractValue.createFromUnaryOp(realm, "!", value, true, value.expressionLocation);
  return _index3.AbstractValue.createFromUnaryOp(realm, "!", v, true, value.expressionLocation);
}

function negate(realm, value) {
  if (value instanceof _index3.ConcreteValue) return _index.ValuesDomain.computeUnary(realm, "!", value);
  (0, _invariant2.default)(value instanceof _index3.AbstractValue);
  if (value.kind === "!") return makeBoolean(realm, value.args[0]);
  // todo #1001: remove this check once intrinsic objects can be properly nullable
  if (!(value instanceof _index3.AbstractObjectValue) || !value.isIntrinsic()) {
    if (!value.mightNotBeTrue()) return realm.intrinsics.false;
    if (!value.mightNotBeFalse()) return realm.intrinsics.true;
  }
  // If NaN is not an issue, invert binary ops
  if (value.args.length === 2 && !value.args[0].mightBeNumber() && !value.args[1].mightBeNumber()) {
    var invertedComparison = void 0;
    switch (value.kind) {
      case "===":
        invertedComparison = "!==";
        break;
      case "==":
        invertedComparison = "!=";
        break;
      case "!==":
        invertedComparison = "===";
        break;
      case "!=":
        invertedComparison = "==";
        break;
      case "<":
        invertedComparison = ">=";
        break;
      case "<=":
        invertedComparison = ">";
        break;
      case ">":
        invertedComparison = "<=";
        break;
      case ">=":
        invertedComparison = "<";
        break;
      default:
        break;
    }
    if (invertedComparison !== undefined) {
      var left = simplify(realm, value.args[0]);
      var right = simplify(realm, value.args[1]);
      return _index3.AbstractValue.createFromBinaryOp(realm, invertedComparison, left, right, value.expressionLocation);
    }
    var invertedLogicalOp = void 0;
    switch (value.kind) {
      case "&&":
        invertedLogicalOp = "||";
        break;
      case "||":
        invertedLogicalOp = "&&";
        break;
      default:
        break;
    }
    if (invertedLogicalOp !== undefined) {
      var _left = negate(realm, value.args[0]);
      var _right = negate(realm, value.args[1]);
      return _index3.AbstractValue.createFromLogicalOp(realm, invertedLogicalOp, _left, _right, value.expressionLocation);
    }
  }
  return _index3.AbstractValue.createFromUnaryOp(realm, "!", value, true, value.expressionLocation);
}
//# sourceMappingURL=simplifier.js.map